/*
 *  Copyright (C) 2002 Alan Hourihane.  All Rights Reserved.
 *
 *  This is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This software is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this software; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
 *  USA.
 *
 *  Author: Alan Hourihane <alanh@fairlite.demon.co.uk>
 */

#include "rfb.h"

#ifdef HAVE_DMX_CONFIG_H
#include "dmx-config.h"
#endif

#ifdef DMXVNC
#include "../dmx.h"
#include "../dmxcb.h"
#endif

#include <time.h>
#include "fb.h"

#include "dixstruct.h"
#include "compiler.h"

#include "mipointer.h"
#include "mibstore.h"

#include "globals.h"
#define DPMS_SERVER
#include <X11/extensions/dpms.h>

#include <netinet/in.h>

int vncScreenPrivateIndex = -1;
int rfbGCIndex = -1;
int inetdSock = -1;
Atom VNC_LAST_CLIENT_ID = 0;
Atom VNC_CONNECT = 0;
char *desktopName = "x11";
char rfbThisHost[256];

extern void VncExtensionInit(void);

extern void vncInitMouse(void);
extern void vncInitKeyb(void);
Bool VNCInit(ScreenPtr pScreen, DMXScreenInfo *dmxScreen);

#ifndef XFree86LOADER
static unsigned long VNCGeneration = 0;
#endif
static void rfbWakeupHandler (int i, pointer blockData, unsigned long err, pointer pReadmask);

/*
 * rfbLog prints a time-stamped message to the log file (stderr).
 */

void rfbLog(char *format, ...)
{
    va_list args;
    char buf[256];
    time_t clock;

    va_start(args, format);

    time(&clock);
    strftime(buf, 255, "%d/%m/%Y %H:%M:%S ", localtime(&clock));
    fprintf(stderr, buf);

    vfprintf(stderr, format, args);
    fflush(stderr);

    va_end(args);
}

void rfbLogPerror(char *str)
{
    rfbLog("");
    perror(str);
}


static ScreenPtr TheVNCScreen = NULL;


static void
nopGetImage(DrawablePtr pDrawable, int sx, int sy,
            int w, int h, unsigned int format,
            unsigned long planemask, char *pdstLine)
{
   ScreenPtr pScreen = pDrawable->pScreen;

   (*pScreen->GetImage)(pDrawable,sx,sy,w,h,format,planemask,pdstLine);
}


static void
SetupVNCScreen(void)
{
    vncScreenPtr pScreenPriv;
    VisualPtr v = NULL;
    int i;

    pScreenPriv = xalloc(sizeof(vncScreenRec));
    pScreenPriv->width = dmxGlobalWidth;
    pScreenPriv->height = dmxGlobalHeight;
    pScreenPriv->rfbAuthTries = 0;
    pScreenPriv->rfbAuthTooManyTries = FALSE;
    pScreenPriv->timer = NULL;
    pScreenPriv->udpPort = 0;
    pScreenPriv->rfbListenSock = -1;
    pScreenPriv->udpSock = -1;
    pScreenPriv->udpSockConnected = FALSE;
    pScreenPriv->httpListenSock = -1;
    pScreenPriv->httpSock = -1;
    pScreenPriv->maxFd = 0;
    pScreenPriv->rfbAuthPasswdFile = NULL;
    pScreenPriv->httpDir = NULL;
    pScreenPriv->rfbInstalledColormap = NULL;
    pScreenPriv->interface.s_addr = htonl (INADDR_ANY);

    pScreenPriv->rfbPort = 0;
    pScreenPriv->httpPort = 0;
    pScreenPriv->rfbAuthPasswdFile = NULL;
    pScreenPriv->httpDir = NULL;
    pScreenPriv->rfbAlwaysShared = FALSE;
    pScreenPriv->rfbNeverShared = FALSE;
    pScreenPriv->rfbUserAccept = FALSE;
    pScreenPriv->rfbViewOnly = FALSE;
    pScreenPriv->rfbDontDisconnect = FALSE;
    pScreenPriv->loginAuthEnabled = FALSE;

    /* Find the root window's visual.
     * XXX this might be the wrong info to use below.
     */
    for (i = 0; i < screenInfo.screens[0]->numVisuals; i++) {
        v = screenInfo.screens[0]->visuals + i;
        if (v->vid == screenInfo.screens[0]->rootVisual) {
            break;
        }
    }

    /* XXX is this the right depth and bpp? */
    /* With DMX, screenInfo.screens[0]->rootDepth is 24 and that doesn't
     * work in translate.c!  We're just using 32.
     */
    pScreenPriv->rfbServerFormat.bitsPerPixel = 32;
    pScreenPriv->rfbServerFormat.depth = 32;
    pScreenPriv->rfbServerFormat.bigEndian = 0;
    pScreenPriv->rfbServerFormat.trueColour = (v->class == TrueColor);
    if (pScreenPriv->rfbServerFormat.trueColour) {
	pScreenPriv->rfbServerFormat.redMax = v->redMask >> v->offsetRed;
	pScreenPriv->rfbServerFormat.greenMax = v->greenMask >> v->offsetGreen;
	pScreenPriv->rfbServerFormat.blueMax = v->blueMask >> v->offsetBlue;
	pScreenPriv->rfbServerFormat.redShift = v->offsetRed;
	pScreenPriv->rfbServerFormat.greenShift = v->offsetGreen;
	pScreenPriv->rfbServerFormat.blueShift = v->offsetBlue;
    } else {
	pScreenPriv->rfbServerFormat.redMax
	    = pScreenPriv->rfbServerFormat.greenMax
	    = pScreenPriv->rfbServerFormat.blueMax = 0;
	pScreenPriv->rfbServerFormat.redShift
	    = pScreenPriv->rfbServerFormat.greenShift
	    = pScreenPriv->rfbServerFormat.blueShift = 0;
    }

    /* Allocate/init TheVNCScreen object */
    vncScreenPrivateIndex = 0;

    TheVNCScreen = (ScreenPtr) Xcalloc(1, sizeof(struct _Screen));
    TheVNCScreen->devPrivates = (DevUnion *) Xcalloc(1, sizeof(DevUnion));
    TheVNCScreen->devPrivates[vncScreenPrivateIndex].ptr = pScreenPriv;
    TheVNCScreen->GetImage = nopGetImage;

    rfbInitSockets(TheVNCScreen);
    if (inetdSock == -1)
       httpInitSockets(TheVNCScreen);
}


/*
 * Called by DMX's InitOutput()
 * Will be called for each X server generation.
 */
void
VNCInitForDMX(void)
{
    rfbLog("DMXVNC: Initializing VNC for DMX\n");
    ErrorF("DMXVNC: Initializing VNC for DMX\n");

    if (TheVNCScreen) {
        rfbLog("DMXVNC: New server generation\n");
    }
    else {
        rfbLog("DMXVNC: Set up VNC screen\n");
        SetupVNCScreen();
    }

    /* Reset the input device pointers.  They'll get set later.
     * This allows successful server regeneration.
    */
    vncSetKeyboardDevice(NULL);
    vncSetPointerDevice(NULL);
}


Bool
VNCInit(ScreenPtr pScreen, DMXScreenInfo *pScrn)
{
    VisualPtr visual;
    vncScreenPtr pScreenPriv;
#if 0
    char *interface_str = NULL;
#endif
#ifdef RENDER
    PictureScreenPtr	ps;
#endif

#ifndef XFree86LOADER
    if (VNCGeneration != serverGeneration) {
	VNCGeneration = serverGeneration;
	if ( ((vncScreenPrivateIndex = AllocateScreenPrivateIndex()) < 0) ||
	    ((rfbGCIndex = AllocateGCPrivateIndex()) < 0) )
		return FALSE;
    }
#endif

    if (!AllocateGCPrivate(pScreen, rfbGCIndex, sizeof(rfbGCRec))) {
	ErrorF("VNCInit(): failed to allocate GCIndex\n");
	return FALSE;
    }

    if (!(pScreenPriv = xalloc(sizeof(vncScreenRec))))
	return FALSE;

    pScreen->devPrivates[vncScreenPrivateIndex].ptr = (pointer)pScreenPriv;

    pScreenPriv->rfbAuthTries = 0;
    pScreenPriv->rfbAuthTooManyTries = FALSE;
    pScreenPriv->timer = NULL;
    pScreenPriv->udpPort = 0;
    pScreenPriv->rfbListenSock = -1;
    pScreenPriv->udpSock = -1;
    pScreenPriv->udpSockConnected = FALSE;
    pScreenPriv->httpListenSock = -1;
    pScreenPriv->httpSock = -1;
    pScreenPriv->maxFd = 0;
    pScreenPriv->rfbAuthPasswdFile = NULL;
    pScreenPriv->httpDir = NULL;
    pScreenPriv->rfbInstalledColormap = NULL;
    pScreenPriv->interface.s_addr = htonl (INADDR_ANY);

    pScreenPriv->rfbPort = 0;
    pScreenPriv->httpPort = 0;
    pScreenPriv->rfbAuthPasswdFile = NULL;
    pScreenPriv->httpDir = NULL;
    pScreenPriv->rfbAlwaysShared = FALSE;
    pScreenPriv->rfbNeverShared = FALSE;
    pScreenPriv->rfbUserAccept = FALSE;
    pScreenPriv->rfbViewOnly = FALSE;
    pScreenPriv->rfbDontDisconnect = FALSE;
    pScreenPriv->loginAuthEnabled = FALSE;

#if 0
    if (xf86ReturnOptValBool(options, OPTION_LOCALHOST, FALSE))
	pScreenPriv->interface.s_addr = htonl (INADDR_LOOPBACK);

    interface_str = xf86GetOptValString(options, OPTION_INTERFACE);

    if (interface_str && pScreenPriv->interface.s_addr == htonl(INADDR_ANY)) {
	Bool failed = FALSE;
	struct in_addr got;
	unsigned long octet;
	char *p = interface_str, *end;
	int q;

	for (q = 0; q < 4; q++) {
	    octet = strtoul (p, &end, 10);

	    if (p == end || octet > 255)
		failed = TRUE;

	    if ((q < 3 && *end != '.') ||
	        (q == 3 && *end != '\0'))
		failed = TRUE;

	    got.s_addr = (got.s_addr << 8) | octet;
	    p = end + 1;
	}

	if (!failed)
	    pScreenPriv->interface.s_addr = htonl (got.s_addr);
	else
	    xf86DrvMsg(pScrn->scrnIndex, X_INFO, "VNC interface option malformed, not using.\n");
    }
#endif

    if (!VNC_LAST_CLIENT_ID)
	VNC_LAST_CLIENT_ID = MakeAtom("VNC_LAST_CLIENT_ID",
				  strlen("VNC_LAST_CLIENT_ID"), TRUE);
    if (!VNC_CONNECT)
	VNC_CONNECT = MakeAtom("VNC_CONNECT", strlen("VNC_CONNECT"), TRUE);

    rfbInitSockets(pScreen);
    if (inetdSock == -1)
	httpInitSockets(pScreen);

#ifdef CORBA
    initialiseCORBA(argc, argv, desktopName);
#endif

    pScreenPriv->width = pScrn->scrnWidth;
    pScreenPriv->height = pScrn->scrnHeight;
    pScreenPriv->depth = pScrn->beDepth;
    pScreenPriv->paddedWidthInBytes = PixmapBytePad(pScrn->scrnWidth, pScrn->beDepth);
    pScreenPriv->bitsPerPixel = rfbBitsPerPixel(pScrn->beDepth);
    pScreenPriv->pfbMemory = NULL;
    pScreenPriv->oldpfbMemory = NULL;

    pScreenPriv->cursorIsDrawn = TRUE;
    pScreenPriv->dontSendFramebufferUpdate = FALSE;

    pScreenPriv->CloseScreen = pScreen->CloseScreen;
    pScreenPriv->CreateGC = pScreen->CreateGC;
    pScreenPriv->PaintWindowBackground = pScreen->PaintWindowBackground;
    pScreenPriv->PaintWindowBorder = pScreen->PaintWindowBorder;
    pScreenPriv->CopyWindow = pScreen->CopyWindow;
    pScreenPriv->ClearToBackground = pScreen->ClearToBackground;
    pScreenPriv->RestoreAreas = pScreen->RestoreAreas;
    pScreenPriv->WakeupHandler = pScreen->WakeupHandler;
    pScreenPriv->InstallColormap = pScreen->InstallColormap;
    pScreenPriv->UninstallColormap = pScreen->UninstallColormap;
    pScreenPriv->ListInstalledColormaps = pScreen->ListInstalledColormaps;
    pScreenPriv->StoreColors = pScreen->StoreColors;
#ifdef CHROMIUM
    pScreenPriv->RealizeWindow = pScreen->RealizeWindow;
    pScreenPriv->UnrealizeWindow = pScreen->UnrealizeWindow;
    pScreenPriv->DestroyWindow = pScreen->DestroyWindow;
    pScreenPriv->PositionWindow = pScreen->PositionWindow;
    pScreenPriv->ResizeWindow = pScreen->ResizeWindow;
    pScreenPriv->ClipNotify = pScreen->ClipNotify;
#endif
#ifdef RENDER
    ps = GetPictureScreenIfSet(pScreen);
    if (ps)
	pScreenPriv->Composite = ps->Composite;
#endif
    pScreen->CloseScreen = rfbCloseScreen;
    pScreen->CreateGC = rfbCreateGC;
    pScreen->PaintWindowBackground = rfbPaintWindowBackground;
    pScreen->PaintWindowBorder = rfbPaintWindowBorder;
    pScreen->CopyWindow = rfbCopyWindow;
    pScreen->ClearToBackground = rfbClearToBackground;
    pScreen->RestoreAreas = rfbRestoreAreas;
    pScreen->WakeupHandler = rfbWakeupHandler;
    pScreen->InstallColormap = rfbInstallColormap;
    pScreen->UninstallColormap = rfbUninstallColormap;
    pScreen->ListInstalledColormaps = rfbListInstalledColormaps;
    pScreen->StoreColors = rfbStoreColors;

#ifdef CHROMIUM
    pScreen->RealizeWindow = rfbRealizeWindow;
    pScreen->UnrealizeWindow = rfbUnrealizeWindow;
    pScreen->DestroyWindow = rfbDestroyWindow;
    pScreen->PositionWindow = rfbPositionWindow;
    pScreen->ResizeWindow = rfbResizeWindow;
    pScreen->ClipNotify = rfbClipNotify;
#endif
#ifdef RENDER
    if (ps)
	ps->Composite = rfbComposite;
#endif

    for (visual = pScreen->visuals; visual->vid != pScreen->rootVisual; visual++)
	;

    if (!visual) {
	ErrorF("rfbScreenInit: couldn't find root visual\n");
	return FALSE;
    }

    pScreenPriv->rfbServerFormat.bitsPerPixel = pScrn->beBPP;
    pScreenPriv->rfbServerFormat.depth = pScrn->beDepth;
    pScreenPriv->rfbServerFormat.bigEndian = !(*(char *)&rfbEndianTest);
    pScreenPriv->rfbServerFormat.trueColour = (visual->class == TrueColor);
    if (pScreenPriv->rfbServerFormat.trueColour) {
	pScreenPriv->rfbServerFormat.redMax = visual->redMask >> visual->offsetRed;
	pScreenPriv->rfbServerFormat.greenMax = visual->greenMask >> visual->offsetGreen;
	pScreenPriv->rfbServerFormat.blueMax = visual->blueMask >> visual->offsetBlue;
	pScreenPriv->rfbServerFormat.redShift = visual->offsetRed;
	pScreenPriv->rfbServerFormat.greenShift = visual->offsetGreen;
	pScreenPriv->rfbServerFormat.blueShift = visual->offsetBlue;
    } else {
	pScreenPriv->rfbServerFormat.redMax
	    = pScreenPriv->rfbServerFormat.greenMax
	    = pScreenPriv->rfbServerFormat.blueMax = 0;
	pScreenPriv->rfbServerFormat.redShift
	    = pScreenPriv->rfbServerFormat.greenShift
	    = pScreenPriv->rfbServerFormat.blueShift = 0;
    }

    return TRUE;
}

static void
rfbWakeupHandler (
    int 	i,
    pointer     blockData,
    unsigned long err,
    pointer     pReadmask
){
    ScreenPtr      pScreen = screenInfo.screens[i];
    vncScreenPtr   pScreenPriv = VNCPTR(pScreen);
    /*DMXScreenInfo *pScrn = &dmxScreens[pScreen->myNum];*/

    rfbRootPropertyChange(pScreen); /* Check clipboard */

    rfbCheckFds(pScreen);
    httpCheckFds(pScreen);
#ifdef CORBA
    corbaCheckFds();
#endif

    pScreen->WakeupHandler = pScreenPriv->WakeupHandler;
    (*pScreen->WakeupHandler) (i, blockData, err, pReadmask);
    pScreen->WakeupHandler = rfbWakeupHandler;
}


/**
 * The rfb wake-up handler used when we're compiled into the DMX server.
 */
void
rfbWakeupHandlerDMX(void)
{
    ScreenPtr      pScreen = TheVNCScreen;
    rfbRootPropertyChange(pScreen); /* Check clipboard */
    rfbCheckFds(pScreen);
    httpCheckFds(pScreen);
}
